/*
 * Modem.java
 *
 * Copyright (C) 2011 John Douyere (VK2ETA) - for Android platforms
 * Copyright (C) 2008 P�r Crusefalk and Rein Couperus
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.AndFlmsg;

import android.content.*;
import android.media.*;
import android.media.MediaRecorder.*;
import java.util.regex.*;



public class Modem {


	public static final int RXMODEMIDLE = 0;
	public static final int RXMODEMSTARTING = 1;
	public static final int RXMODEMRUNNING = 2;
	public static final int RXMODEMPAUSED = 3;
	public static final int RXMODEMSTOPPING = 4;

	public int modemState = RXMODEMIDLE;
	
	public boolean modemThreadOn = true;
	
//	public static boolean rxHasStopped = true;
	public static int NumberOfOverruns = 0;
	private static AudioRecord audiorecorder = null;
	private boolean RxON = false;
	private int bufferSize = 0;
	private final float sampleRate = 8000.0f;


	private static boolean FirstBracketReceived = false;
	public static String MonitorString = "";
	public static boolean WrapLF = false;
	private static String BlockString = "";
	private static int RxBlockSize = 0;
	

	public static TxPSK TXMpsk;
	public static TxMFSK TXMmfsk;
	public static TxTHOR TXMthor;
	public static RxPSK RXMpsk;
	public static RxMFSK RXMmfsk;
	public static RxTHOR RXMthor;
	public static RxRSID myRxRSID;
	
	private static SampleRateConversion myResampler;

	public double rxFrequency = 1500.0; 
    double squelch = 20.0;

	public static boolean stopTX = false;


	public Modem () {

		//Prepare RSID Receive
		myRxRSID =  new RxRSID();
		//Initialise modems
		RXMpsk = new RxPSK(Processor.RxModem);
		RXMmfsk = new RxMFSK(Processor.RxModem);
		RXMthor = new RxTHOR(Processor.RxModem);
		//Initialize Re-sampling to 11025Hz for RSID, THOR and MFSK modems
		myResampler = new SampleRateConversion(11025.0 / 8000.0);

	}

	public static String getModemString (modemmodeenum mode) {
		return mode.toString();
	}


	public static String getAltModemString (modemmodeenum mode) {
		return mode.toString();
	}

	//THE ORDER MUST MATCH the list in modemmodeenum.java
	static public final char modemmodecode[] = { 'f','1','2','d','g','e','3','4','b','a','5','9','c','6','8','7' };
	static public final String modemmodecodeSTR = "786c95ab43d21";

	public static modemmodeenum getmode (char modemcode) {
		modemmodeenum outmode = modemmodeenum.PSK250R;
		switch (modemcode){
		case 'f':
			outmode = modemmodeenum.MFSK8;
			break;
		case '1':
			outmode = modemmodeenum.THOR8;
			break;
		case '2':
			outmode = modemmodeenum.MFSK16;
			break;
		case 'd':
			outmode = modemmodeenum.THOR11;
			break;
		case 'g':
			outmode = modemmodeenum.THOR16;
			break;
		case 'e':
			outmode = modemmodeenum.PSK31;
			break;
		case '3':
			outmode = modemmodeenum.THOR22;
			break;
		case '4':
			outmode = modemmodeenum.MFSK32;
			break;
		case 'b':
			outmode = modemmodeenum.PSK125R;
			break;
		case 'a':
			outmode = modemmodeenum.PSK63;
			break;
		case '5':
			outmode = modemmodeenum.PSK250R;
			break;
		case '9':
			outmode = modemmodeenum.PSK125;
			break;
		case 'c':
			outmode = modemmodeenum.MFSK64;
			break;
		case '6':
			outmode = modemmodeenum.PSK500R;
			break;
		case '8':
			outmode = modemmodeenum.PSK250;
			break;
		case '7':
			outmode = modemmodeenum.PSK500;
			break;
		default:
			outmode = modemmodeenum.PSK250R;                                 
		}                  
		return outmode;
	}


	private void soundInInit() {

		bufferSize = (int) sampleRate; // 1 second of Audio max
		if (bufferSize < AudioRecord.getMinBufferSize((int) sampleRate, AudioFormat.CHANNEL_IN_MONO , AudioFormat.ENCODING_PCM_16BIT)) { 
			// Check to make sure buffer size is not smaller than the smallest allowed one 
			bufferSize = AudioRecord.getMinBufferSize((int) sampleRate, AudioFormat.CHANNEL_IN_MONO , AudioFormat.ENCODING_PCM_16BIT);
		}
		int ii = 20; //number of 1/4 seconds wait
		while (--ii > 0) {
			if (AndFlmsg.toBluetooth) {
				//Bluetooth hack (use voice call)
				audiorecorder = new AudioRecord(AudioSource.MIC, 8000, android.media.AudioFormat.CHANNEL_IN_MONO, 
						android.media.AudioFormat.ENCODING_PCM_16BIT, bufferSize);
			} else {
				audiorecorder = new AudioRecord(AudioSource.MIC, 8000, android.media.AudioFormat.CHANNEL_IN_MONO, 
						android.media.AudioFormat.ENCODING_PCM_16BIT, bufferSize);
			}
			if (audiorecorder.getState() == AudioRecord.STATE_INITIALIZED) {
				ii = 0;//ok done
			} else {
				if (ii < 16) { //Only if have to wait more than 1 seconds
					loggingclass.writelog("Waiting for Audio MIC availability..." , null, true);
				}
				try {
					Thread.sleep(250);//1/4 second
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		if (audiorecorder.getState() != AudioRecord.STATE_INITIALIZED) {
			//JD add exception catch here
			System.out.println("Can't open audio MIC");
			loggingclass.writelog("Can't open Audio MIC " , null, true);
		}
	}



	private void rxInit() {
			//Initialise PSK modems
			if (RXMpsk != null) RXMpsk.rxInit();
			//Initialise PSK modems
			if (RXMmfsk != null) RXMmfsk.rxInit();
			//Initialise PSK modems
			if (RXMthor != null) RXMthor.rxInit();
	}


	public void startmodem() {
		
		modemThreadOn = true;

		new Thread(new Runnable() {
			public void run() {


				while (modemThreadOn) {


					modemState = RXMODEMSTARTING;
					double startproctime = 0;
					double endproctime = 0;
					int numSamples8K = 0;
					soundInInit();
					rxInit();
					NumberOfOverruns = 0;
					audiorecorder.startRecording();
					RxON = true;
					Processor.restartRxModem.drainPermits();
					//Since the callback is not working, implement a while loop.
					short[] so8K = new short[bufferSize];
					int size12Kbuf =  (int) ((bufferSize + 1) * 11025.0 / 8000.0);
					double[] so12K = new double[size12Kbuf];
					if (Processor.RxModem.toString().startsWith("PSK")) {
						//Change PSK modem
						Modem.RXMpsk.changemode(Processor.RxModem);
					} else if (Processor.RxModem.toString().startsWith("THOR")) {
						//Change MFSK modem
						Modem.RXMthor.changemode(Processor.RxModem);
					} else if (Processor.RxModem.toString().startsWith("MFSK")) {
						//Change MFSK modem
						Modem.RXMmfsk.changemode(Processor.RxModem);
					} else {
						loggingclass.writelog("Wrong RX Mode called in RxThread: " + Processor.TxModem.toString() , null, true);
					}
					while (RxON) {
						endproctime = System.currentTimeMillis();
						double buffertime = (double) numSamples8K / 8000.0 * 1000.0; //in milliseconds
						if (numSamples8K > 0) Processor.cpuload = (int)(((double)(endproctime - startproctime)) / buffertime * 100);
						if (Processor.cpuload > 100) Processor.cpuload = 100;
						AndFlmsg.mHandler.post(AndFlmsg.updatecpuload);
						numSamples8K = audiorecorder.read(so8K, 0, 8000/4); //process only part of the buffer to avoid lumpy processing
						if (numSamples8K > 0) {
							modemState = RXMODEMRUNNING;
							startproctime = System.currentTimeMillis();
							//Process only if Rx is ON, otherwise discard (we have already decided to TX)
							if (RxON) {
								//Re-sample to 11025Hz for RSID, THOR and MFSK modems
								int numSamples12K = myResampler.Process(so8K, numSamples8K, so12K, size12Kbuf);
								//RxRSID always ON
								myRxRSID.receive(so12K, numSamples12K);
								//Then the selected RX modem
								if (Processor.RxModem.toString().startsWith("PSK")) {
									RXMpsk.rxProcess(so8K, numSamples8K);
								} else if (Processor.RxModem.toString().startsWith("MFSK")) {
									RXMmfsk.rxProcess(so8K, numSamples8K);
								} else if (Processor.RxModem.toString().startsWith("THOR")) {
									RXMthor.rxProcess(so8K, numSamples8K, so12K, numSamples12K);
								} else {
									loggingclass.writelog("Wrong RX Mode called: " + Processor.RxModem.toString() , null, true);
								}
							}
							//Post to monitor (Modem) window after each buffer processing
							//Add TX frame too if present
							if (Modem.MonitorString.length() > 0 || Processor.TXmonitor.length() > 0) {
								Processor.monitor += Modem.MonitorString + Processor.TXmonitor;
								Processor.TXmonitor = "";
								Modem.MonitorString = "";
								AndFlmsg.mHandler.post(AndFlmsg.addtomodem);
							}
						}
					}//while (RxON)
					//We dropped here on pause flag
					if (audiorecorder != null) {
						//Avoid some crashes on wrong state
						if (audiorecorder.getState() == AudioRecord.STATE_INITIALIZED) {
							if (audiorecorder.getRecordingState() == AudioRecord.RECORDSTATE_RECORDING) {
								audiorecorder.stop();
							}
							audiorecorder.release();
						}
					}
					modemState = RXMODEMPAUSED;
					//Marker for end of thread (Stop modem thread flag)
					if (!modemThreadOn) {
						modemState = RXMODEMIDLE;
						return;
					}
					//Now waits for a restart (or having this thread killed)
					Processor.restartRxModem.acquireUninterruptibly(1);
					//Make sure we don's have spare permits
					Processor.restartRxModem.drainPermits();
				}//while (modemThreadOn)
				//We dropped here on thread stop request
				modemState = RXMODEMIDLE;				
			} //run
		}).start(); //	new Thread(new Runnable() {

	}


	public void stopRxModem() {
		modemThreadOn = false;
		RxON = false;
	}

	public void pauseRxModem() {
		RxON = false;
	}

	public void unPauseRxModem() {
    	Processor.restartRxModem.release(1);
	}

	void changemode(modemmodeenum newmode) {
		//Stop the modem receiving side to prevent using the wrong values
		pauseRxModem();
		//Restart modem reception
		unPauseRxModem();

	}

	void setFrequency(double rxfreq) {
		rxFrequency = rxfreq;
	}
	
	void reset() {
		String frequencySTR = AndFlmsg.myconfig.getPreference("AFREQUENCY","1000");
		rxFrequency = Integer.parseInt(frequencySTR);
		if (rxFrequency < 500) rxFrequency = 500;
		if (rxFrequency > 2500) rxFrequency = 2500;
		squelch = AndFlmsg.mysp.getFloat("SQUELCHVALUE", (float) 20.0);
	}
	
	
    /**
     * @param squelchdiff the delta to add to squelch
     */
    public void AddtoSquelch(double squelchdiff) {
        squelch += (squelch > 10) ? squelchdiff : squelchdiff / 2;
        if (squelch < 0) squelch =0;
        if (squelch >100) squelch =100;
		//store value into preferences
	      SharedPreferences.Editor editor = AndFlmsg.mysp.edit();
	      editor.putFloat("SQUELCHVALUE", (float) squelch);
	      // Commit the edits!
	      editor.commit();

    }

	public static void rxblock (char inChar) {
	
		//No TX yet
		//if (!Processor.ReceivingForm) {
		//	Processor.DCD = Processor.MAXDCD;
		//}

		if (inChar > 127) {
			// todo: unicode encoding
			inChar = 0;
		}
		
		switch (inChar) {
			case 0:
				break; // do nothing
			case '[':
				if (Processor.ReceivingForm ||
					FirstBracketReceived) {
					BlockString += inChar;
				} else {			
					if (!FirstBracketReceived) {
						FirstBracketReceived = true;
						BlockString = "[";
						//Processor.PostToTerminal("FirstBracket\n");
					} 
				}
				//	            Main.DCD = 0;
				break;
			case ']':
				if (Processor.ReceivingForm) {
					BlockString += inChar;
					//Did we receive a new start of block before the end 
					//of the previous block? If so, restart reception
					if (BlockString.endsWith("[WRAP:beg]")) {
						BlockString = "[WRAP:beg]";
						Processor.PostToTerminal("\n\nReceiving New Message, discarding any partially received data\n");
					} else if (BlockString.endsWith("[WRAP:lf]")) {
						WrapLF = true;
					} else if (BlockString.endsWith("[WRAP:crlf]")) {
							WrapLF = false;
					} else if (BlockString.endsWith("[WRAP:end]")) {
						//We have a complete block, process it
						//Processor.PostToTerminal("End\n");
						Processor.ProcessBlock(BlockString);
						Processor.CrcString = "";
						Processor.FileNameString = "";
						Processor.ReceivingForm = false;
						FirstBracketReceived = false;
					} else {
						//Check for CRC section
						Pattern psc = Pattern.compile("WRAP:chksum ([0-9A-F]{4})");
						Matcher msc = psc.matcher(BlockString);
						if (msc.find()) {
							Processor.CrcString = msc.group(1);
							//Processor.PostToTerminal("Checksum <" + Processor.CrcString + ">\n");
						}
						//Check for file name section
						psc = Pattern.compile("\\[WRAP:fn (.+\\..+)\\]");
						msc = psc.matcher(BlockString);
						if (msc.find()) {
							if (BlockString.endsWith(msc.group(0))) {
								Processor.FileNameString = msc.group(1);
								Processor.PostToTerminal("\nReceiving File " + Processor.FileNameString);
								//Processor.PostToTerminal("\n" + msc.group(0) + "\n");
							}
						}
					}
				} else {
					BlockString += inChar;
					if (BlockString.endsWith("[WRAP:beg]")) {
						FirstBracketReceived = true;
						BlockString = "[WRAP:beg]";
						Processor.ReceivingForm = true;
						Processor.PostToTerminal("\n\nReceiving New Message");
					}

				}
				break;
			case 10:
				MonitorString += "\n";
				if (Processor.ReceivingForm || 
					FirstBracketReceived) {
					BlockString += inChar;
				}
				break;
			case 13: //ignore CR
				break;
			default:
				if (Processor.ReceivingForm || 
					FirstBracketReceived) {
					BlockString += inChar;
				}
				break;
		}	// end switch
		//Reset if header not found within
		//  1000 charaters of first bracket
		if (FirstBracketReceived && 
			!Processor.ReceivingForm &&
			BlockString.length() > 1000) {
			BlockString = BlockString.substring(900);
			if (!BlockString.contains("]"))  FirstBracketReceived = false;
		}
		//Reset if end marker not found within
		// 30,000 charaters of start
		//to-do: if found crc but not wrap end, process anyway
		if (Processor.ReceivingForm &&
			BlockString.length() > 30000)
		{
			Processor.CrcString = "";
			Processor.FileNameString = "";
			Processor.ReceivingForm = false;
			FirstBracketReceived = false;	
		}
		if (inChar > 31)
		{
			MonitorString += inChar;
		}
	}


	//In a separate thread so that the UI thread is not blocked during TX
	public void sendln(String Sendline)
	{
		Runnable TxRun = new TxThread(Sendline);
		Sendline = "";
		new Thread(TxRun).start();
	}

	public class TxThread implements Runnable {
		private String TxSendline = "";

		public TxThread(String Sendline) {
			TxSendline = Sendline;
		}

		public void run() {
			// see if tx active and DCD is off
			//JD Debug - FIX this: review DCD logic        if (Sendline.length() > 0 & !Processor.TXActive & Processor.DCD == 0) {
			if (TxSendline.length() > 0 & !Processor.TXActive) {

				//JD check what this is for?
				//        	DCDthrow = generator.nextInt(Persistence);

				if (Processor.ReceivingForm | Processor.Connecting | Processor.Aborting){
					if (Processor.Aborting) {
						Processor.Aborting = false;
					}
					Processor.DCDthrow = 0;
				}
				if (Processor.DCDthrow == 0){
					try {
						Processor.TXActive = true;
						AndFlmsg.mHandler.post(AndFlmsg.updatetitle);   
						//Stop the modem receiving side
						pauseRxModem();
						//TX delay here from config popup (allows on-the-fly changes)
						int TxDelay = 0;
						TxDelay = AndFlmsg.myconfig.getTxDelay();

						if (TxDelay > 0) {
							Thread.sleep(TxDelay);
						}

						/* not yet                    //  Add a 2 seconds delay when mode is MFSK16 (1 sec for MFSK32) to prevent overlaps as
                    //  the trail of MFSK is very long
                    if (Main.RxModem.equals(modemmodeenum.MFSK16)) {
                        Thread.sleep(2000);
                    } else if (Main.RxModem.equals(modemmodeenum.MFSK32)) {
                        Thread.sleep(1000);
                    }
						 */

						//Reset receive marker of RSID for next RX
						Processor.justReceivedRSID = false;

						if (Processor.TxModem.toString().startsWith("PSK")) {
							//Initialise PSK modems
							TXMpsk = new TxPSK(Processor.TxModem);
						} else if (Processor.TxModem.toString().startsWith("MFSK")) {
							//Initialise MFSK modems
							TXMmfsk = new TxMFSK(Processor.TxModem);
						} else if (Processor.TxModem.toString().startsWith("THOR")) {
							//Initialise THOR modems
							TXMthor = new TxTHOR(Processor.TxModem);
						} else {
							loggingclass.writelog("Wrong TX Mode called: " + Processor.TxModem.toString() , null, true);
						}

						//Send TX RSID if required
						if (Processor.TXID) {
							//Reset the send RSID flag
				    		Processor.TXID = false; 
							TxRSID.send();
						}

						if (Processor.TxModem.toString().startsWith("PSK")) {
							TXMpsk.AddBytes(TxSendline.getBytes());
						} else if (Processor.TxModem.toString().startsWith("MFSK")) {
							TXMmfsk.AddBytes(TxSendline.getBytes());
						} else if (Processor.TxModem.toString().startsWith("THOR")) {
							TXMthor.AddBytes(TxSendline.getBytes());
						} else {
							loggingclass.writelog("Wrong TX Mode called: " + Processor.TxModem.toString() , null, true);
						}
						//Restart modem reception
						unPauseRxModem();
						Processor.TXActive = false;
						AndFlmsg.mHandler.post(AndFlmsg.updatetitle);   
					}
					catch (Exception e) {
						loggingclass.writelog("Can't output sound. Is Sound device busy?", null, true);
					}

				}

			}
		}
	};




	//Send Tune in a separate thread so that the UI thread is not blocked 
	//  during TX
	public void TxTune() {
		Runnable txTuneThread = new TxTuneThread();
		new Thread(txTuneThread).start();
	}

	private class TxTuneThread implements Runnable {
		private AudioTrack at = null; 


		public TxTuneThread() {
		}

		public void run() {
			// Check that we are not TXing

			//Stop the modem receiving side
			pauseRxModem();

			//Wait 1/2 second so that if there is potential RF feedback 
			//  on the touchscreen we do not start TXing while the 
			//  finger is still on the screen
			try {
				Thread.sleep(500);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			Processor.TXActive = true;
			String frequencySTR = AndFlmsg.myconfig.getPreference("AFREQUENCY","1000");
			int frequency = Integer.parseInt(frequencySTR);

			int volumebits = Integer.parseInt(AndFlmsg.myconfig.getPreference("VOLUME","8"));

			//Note the multiplier value for the buffer size
			int intSize = 4 * android.media.AudioTrack.getMinBufferSize(8000,
					AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT);

			if (AndFlmsg.toBluetooth) {
				//JD Bluetooth hack test
				//		        	at = new AudioTrack(AudioManager.STREAM_MUSIC, 8000, AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT, intSize , AudioTrack.MODE_STREAM);
				at = new AudioTrack(AudioManager.STREAM_VOICE_CALL, 8000, AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT, intSize , AudioTrack.MODE_STREAM);
			} else {
				at = new AudioTrack(AudioManager.STREAM_MUSIC, 8000, AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT, intSize , AudioTrack.MODE_STREAM);
			}

			//Open audiotrack
			at.setStereoVolume(1.0f,1.0f);
			at.play();

			int sr = 8000; // should be active_modem->get_samplerate();
			int symlen = (int) (1 *  sr); //1 second buffer
			short[] outbuf = new short[symlen];

			double phaseincr;
			double phase = 0.0;
			phaseincr = 2.0 * Math.PI * frequency / sr;

			for (int i = 0; i < 3; i++) { //3 seconds tune
				for (int j = 0; j < symlen; j++) {
					phase += phaseincr;
					if (phase > 2.0 * Math.PI) phase -= 2.0 * Math.PI;
					outbuf[j] = (short) ((int) (Math.sin(phase) * 8386560) >> volumebits);
				}
				at.write(outbuf, 0, symlen);
			}

			//Stop audio track
			at.stop();
			//Wait for end of audio play to avoid 
			//overlaps between end of TX and start of RX
			while (at.getPlayState() == AudioTrack.PLAYSTATE_PLAYING) {
				try {
					Thread.sleep(50);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
	//				e.printStackTrace();
				}
			}
			//Close audio track
			at.release();							

			Processor.TXActive = false;

			//Restart the modem receiving side
			unPauseRxModem();

		}
	};


}
